import re

from letterboxdpy.utils.utils_file import JsonFile
from letterboxdpy.utils.utils_transform import month_to_index
from letterboxdpy.utils.utils_parser import get_meta_content, get_body_content
from letterboxdpy.core.scraper import parse_url
from letterboxdpy.avatar import Avatar
from letterboxdpy.utils.utils_url import extract_path_segment
from letterboxdpy.constants.project import DOMAIN


class UserProfile:

    def __init__(self, username: str) -> None:
        self.username = username
        self.url = f"{DOMAIN}/{self.username}"
        self.dom = parse_url(self.url)

    def __str__(self) -> str:
        return f"Not printable object of type: {self.__class__.__name__}"

    def get_id(self) -> int: return extract_id(self.dom)
    def get_hq_status(self) -> bool: return extract_hq_status(self.dom)
    def get_display_name(self) -> str: return extract_display_name(self.dom)
    def get_bio(self) -> str | None: return extract_bio(self.dom)
    def get_location(self) -> str | None: return extract_location(self.dom)
    def get_website(self) -> str | None: return extract_website(self.dom)
    def get_watchlist_length(self) -> int | None: return extract_watchlist_length(self.dom)
    def get_stats(self) -> dict: return extract_stats(self.dom)
    def get_favorites(self) -> dict: return extract_favorites(self.dom)
    def get_avatar(self) -> dict | None: return extract_avatar(self.dom)
    def get_watchlist_recent(self) -> dict: return extract_watchlist_recent(self.dom)
    def get_diary_recent(self) -> dict: return extract_diary_recent(self.dom)


def extract_id(dom) -> int:
    """Extracts the user ID from the DOM and returns it."""
    try:
        # Alternative method using the report button data attribute
        button_report = dom.find("button", {"data-js-trigger": "report"})
        if button_report:
            report_url = button_report['data-report-url']
            return int(extract_path_segment(report_url, after='person:', before='/'))
        else:
            # Fallback method using regex if button is not found
            pattern = r"/ajax/person:(\d+)/report-for"
            match = re.search(pattern, dom.prettify())
            if match:
                return int(match.group(1))
            else:
                raise ValueError("User ID not found in the DOM")
    except (AttributeError, ValueError) as e:
        raise RuntimeError("Failed to extract user ID from DOM") from e

def extract_hq_status(dom) -> bool:
    """Checks if the user has a high-quality account and returns the status."""
    try:
        data = dom.find("div", {'class': 'profile-summary'})
        if data and 'data-profile-summary-options' in data.attrs:
            profile_summary = data['data-profile-summary-options']
            return JsonFile.parse(profile_summary)['isHQ']
        else:
            return 'profile-hq' in (get_body_content(dom, 'class') or [])
    except Exception as e:
        raise RuntimeError("Failed to check HQ status from DOM") from e

def extract_display_name(dom) -> str:
    """Extracts the display name from the DOM using the Open Graph title property."""
    try:
        content = get_meta_content(dom, property='og:title')
        if content:
            return content[:-10]  # Remove the last 10 characters (e.g., " | Letterboxd")
        else:
            raise ValueError("Display name not found in the DOM")
    except Exception as e:
        raise RuntimeError("Failed to extract display name from DOM") from e

def extract_bio(dom) -> str | None:
    """Extracts the bio from the DOM using the Open Graph description property."""
    try:
        content = get_meta_content(dom, property='og:description')
        if content:
            return content.split('Bio: ')[-1] if 'Bio: ' in content else None
        else:
            raise ValueError("Bio not found in the DOM")
    except Exception as e:
        raise RuntimeError("Failed to extract bio from DOM") from e

def extract_location(dom) -> str | None:
    """Extracts the location from the DOM and returns it."""
    try:
        data = dom.find("div", {"class": ["profile-metadata"]})
        if data:
            location = data.find("div", {"class": ["metadatum"]})
            return location.find("span").text if location else None
        return None
    except Exception as e:
        raise RuntimeError("Failed to extract location from DOM") from e

def extract_website(dom) -> str | None:
    """Extracts the website from the DOM."""
    try:
        data = dom.find("div", {"class": ["profile-metadata"]})
        if data:
            website = data.find("a")
            return website['href'] if website else None
        return None
    except Exception as e:
        raise RuntimeError("Failed to extract website from DOM") from e

def extract_watchlist_length(dom) -> int | None:
    """Extracts the watchlist length from the DOM."""
    try:
        nav_links = dom.find_all("a", {"class": ["navlink"]})
        for link in nav_links:
            if "Watchlist" in link.text:
                if "rel" in link.attrs:
                    # 'User watchlist is not visible'
                    return None  # feature: PrivateData(Exception)
                else:
                    # 'User watchlist is visible'
                    widget = dom.find("section", {"class": ["watchlist-aside"]})
                    return int(
                        widget.find("a", {"class": ["all-link"]}).text.replace(',', '')
                    ) if widget else 0
        else:
            # HQ members
            # If no matching link found, return None
            return None
    except Exception as e:
        raise RuntimeError("Failed to extract watchlist length from DOM") from e

def extract_stats(dom) -> dict:
    """Extracts user statistics from the parsed DOM."""
    def extract_stat(stat_element) -> tuple:
        """Extracts the key-value pair from a stat element."""
        value = stat_element.span.text
        key = stat_element.text.lower().replace(value, '').replace(' ', '_').strip()
        return key, int(value.replace(',', ''))

    try:
        stats_elements = dom.find_all("h4", {"class": ["profile-statistic"]})
        if not stats_elements:
            raise ValueError("No statistics found in the DOM")

        return dict(extract_stat(stat) for stat in stats_elements)
    except AttributeError as e:
        raise RuntimeError("Error while parsing DOM structure") from e

def extract_favorites(dom) -> dict:
    """Extracts favorite films from the DOM and returns them as a dictionary."""
    try:
        favorites_section = dom.find("section", {"id": "favourites"})
        if not favorites_section:
            return {}
        
        poster_list = favorites_section.find("ul", class_="poster-list")
        if not poster_list:
            return {}
            
        favorites = {}
        items = poster_list.find_all("li")
        
        for item in items:
            react_div = item.find("div", class_="react-component")
            if not react_div:
                continue
                
            # Extract data from react component attributes
            movie_id = react_div.get('data-film-id')
            movie_slug = react_div.get('data-item-slug')
            movie_name = react_div.get('data-item-name')
            
            # Clean movie name (remove year) like we did in diary
            if movie_name and ' (' in movie_name and movie_name.endswith(')'):
                movie_name = movie_name.rsplit(' (', 1)[0]
            
            if movie_id and movie_slug and movie_name:
                # Extract additional data
                full_display_name = react_div.get('data-item-full-display-name', '')
                target_link = react_div.get('data-target-link', '')
                
                # Extract release year from full display name
                release_year = None
                if full_display_name and '(' in full_display_name and ')' in full_display_name:
                    try:
                        year_part = full_display_name.split('(')[-1].split(')')[0]
                        release_year = int(year_part) if year_part.isdigit() else None
                    except (ValueError, IndexError):
                        pass
                
                favorites[movie_id] = {
                    'slug': movie_slug,
                    'name': movie_name,
                    'url': f'https://letterboxd.com/film/{movie_slug}/',
                    'year': release_year,
                    'log_url': f'https://letterboxd.com{target_link}' if target_link else None
                }
        
        return favorites
    except Exception as e:
        raise RuntimeError("Failed to extract favorites from DOM") from e

def extract_avatar(dom) -> dict | None:
    """Extracts the user's avatar URL from the DOM and returns it."""
    try:
        elem_avatar = dom.find("div", {"class": ["profile-avatar"]})
        if elem_avatar and elem_avatar.img:
            avatar_url = elem_avatar.img['src']
            return Avatar(avatar_url).upscaled_data
        return None  # Avatar not found
    except Exception as e:
        raise RuntimeError("Failed to extract avatar from DOM") from e

def extract_watchlist_recent(dom) -> dict:
    """Extracts recent watchlist items from the DOM, with error handling."""
    def extract_movie_info(item) -> dict:
        """Extracts movie information from a watchlist item."""
        from letterboxdpy.utils.utils_string import extract_year_from_movie_name, clean_movie_name
        
        # Look for data attributes in the nested react-component div
        react_div = item.find('div', {'class': 'react-component'})
        
        if react_div:
            film_id = react_div.get('data-film-id')
            film_slug = react_div.get('data-item-slug')
            raw_name = react_div.get('data-item-name', "Unknown")
            film_name = clean_movie_name(raw_name)
            year = extract_year_from_movie_name(raw_name)
        else:
            # Fallback to original method (for backwards compatibility)
            film_id = item.get('data-film-id')
            film_slug = item.get('data-film-slug')
            raw_name = item.img.get('alt', "Unknown") if item.img else "Unknown"
            film_name = clean_movie_name(raw_name)
            year = extract_year_from_movie_name(raw_name)

        return {
            'id': film_id,
            'slug': film_slug,
            'name': film_name,
            'year': year
        }

    watchlist_recent = {}
    section = dom.find("section", {"class": ["watchlist-aside"]})

    if not section:
        # User watchlist is not visible or there are no items in the watchlist
        return watchlist_recent

    watchlist_items = section.find_all("li", {"class": ["posteritem"]})

    if not watchlist_items:
        raise ValueError("No watchlist items found in the DOM")

    for item in watchlist_items:
        movie = extract_movie_info(item)
        if movie['id']:
            watchlist_recent[movie['id']] = movie

    return watchlist_recent

def extract_diary_recent(dom) -> dict:
    """Extracts recent diary entries from the DOM."""
    diary_recent = {'months':{}}
    # diary
    sections = dom.find_all("section", {"class": ["section"]})
    for section in sections:
        if section.h2 is None:
            continue

        if section.h2.text == "Diary":
            entry_list = section.find_all("li", {"class": ["listitem"]})

            for entry in entry_list:
                month_index = month_to_index(entry.h3.text)

                if month_index not in diary_recent['months']:
                    diary_recent['months'][month_index] = {}

                days = entry.find_all("dt")
                items = entry.find_all("dd")

                for day, item in zip(days, items):
                    movie_index = day.text
                    movie_slug = extract_path_segment(item.a['href'], after='/film/')
                    movie_name = item.text 

                    if movie_index not in diary_recent['months'][month_index]:
                        diary_recent['months'][month_index][movie_index] = []

                    diary_recent['months'][month_index][movie_index].append({
                        'name': movie_name,
                        'slug': movie_slug
                    })
            else:
                break
    return diary_recent



from letterboxdpy.constants.project import MONTH_ABBREVIATIONS

def month_to_index(month_abbreviation):
    """Convert a month abbreviation to its index."""
    try:
        return MONTH_ABBREVIATIONS.index(month_abbreviation) + 1
    except ValueError:
        return None

def index_to_month(month_index):
    """Convert a month index to its abbreviation."""
    if 1 <= month_index <= 12:
        return MONTH_ABBREVIATIONS[month_index - 1]
    return None

def get_ajax_url(url: str) -> str:
    """
    this function returns the ajax url of the given url.
    """
    x = '.com/films'
    ax = ".com/films/ajax"
    
    return url if ax in url else url.replace(x, ax)



from letterboxdpy.core.scraper import Scraper, url_encode
from bs4 import BeautifulSoup
import unittest


class TestScraper(unittest.TestCase):

    def setUp(self):
        self.scraper = Scraper("letterboxd.com")

        self.valid_film_url = "https://letterboxd.com/film/dune-part-two/"
        self.invalid_film_url = "https://letterboxd.com/film/duneparttwo/"

    def test_valid_film_url(self):
        self.assertIsInstance(
            self.scraper.get_page(self.valid_film_url), BeautifulSoup
        )
    
    def test_invalid_film_url(self):
        with self.assertRaises(Exception):
            self.scraper.get_page(self.invalid_film_url)

    def test_url_encode(self):
        query = "Dune: Part Two"
        encoded_query = url_encode(query)
        self.assertEqual(encoded_query, "Dune%3A%20Part%20Two")

if __name__ == '__main__':
    unittest.main()



from typing import TYPE_CHECKING
from curl_cffi import requests
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.core.scraper import Scraper

if TYPE_CHECKING:
    from letterboxdpy.core.models import MovieJSON

class FilmURL:
    """Endpoints and Factory for movie-related data."""
    
    @staticmethod
    def json_url(slug: str) -> str:
        return f"{DOMAIN}/film/{slug}/json/"

    @classmethod
    def json(cls, slug: str) -> 'MovieJSON':
        """Factory: Returns MovieJSON model from film slug."""
        from letterboxdpy.core.models import MovieJSON
        url = cls.json_url(slug)
        
        # Standalone fetch to avoid touching other files
        response = requests.get(url, headers=Scraper.headers, impersonate="chrome")
        if response.status_code != 200:
            from letterboxdpy.core.exceptions import InvalidResponseError
            raise InvalidResponseError(f"Failed to fetch JSON from {url}: {response.status_code}")
            
        return MovieJSON.from_dict(response.json())

    # CSI (Client Side Includes) Endpoints
    @staticmethod
    def _csi(slug: str, endpoint: str) -> str:
        return f"{DOMAIN}/csi/film/{slug}/{endpoint}/"

    @classmethod
    def popular_lists(cls, slug: str) -> str: return cls._csi(slug, "popular-lists")
    @classmethod
    def recent_reviews(cls, slug: str) -> str: return cls._csi(slug, "recent-reviews")
    @classmethod
    def rating_histogram(cls, slug: str) -> str: return cls._csi(slug, "rating-histogram")
    @classmethod
    def user_actions(cls, slug: str) -> str: return cls._csi(slug, "sidebar-user-actions")
    @classmethod
    def stats(cls, slug: str) -> str: return cls._csi(slug, "stats")
    @classmethod
    def news(cls, slug: str) -> str: return cls._csi(slug, "news")
    @classmethod
    def availability(cls, slug: str) -> str: return cls._csi(slug, "availability")
    @classmethod
    def friend_reviews(cls, slug: str) -> str: return cls._csi(slug, "friend-reviews")
    @classmethod
    def friend_activity(cls, slug: str) -> str: return cls._csi(slug, "friend-activity")
    @classmethod
    def own_reviews(cls, slug: str) -> str: return cls._csi(slug, "own-reviews")
    @classmethod
    def liked_reviews(cls, slug: str) -> str: return cls._csi(slug, "liked-reviews")


class UserURL:
    """Endpoints for user-related data."""
    
    @staticmethod
    def homepage() -> str:
        return f"{DOMAIN}/ajax/user-homepage/"

    @staticmethod
    def live_feed() -> str:
        return f"{DOMAIN}/csi/films-live-feed/"


class GeneralURL:
    """General Letterboxd endpoints."""
    
    @staticmethod
    def metadata() -> str:
        return f"{DOMAIN}/ajax/letterboxd-metadata/"



from datetime import datetime

# Date/Time Constants
now = datetime.now()
CURRENT_YEAR = now.year
CURRENT_MONTH = now.month
CURRENT_DAY = now.day
MONTH_ABBREVIATIONS = [
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
]
DAY_ABBREVIATIONS = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

# Domain/URL Constants
URL_PROTOCOLS = ['http://', 'https://']

DOMAIN_FULL = 'letterboxd.com'
DOMAIN_SHORT = 'boxd.it'

# Base URLs
DOMAIN = f'https://{DOMAIN_FULL}'
SITE = f'{DOMAIN}/'
SITE_SHORT = f'https://{DOMAIN_SHORT}/'

DOMAIN_MATCHES = [f'{DOMAIN_FULL}/', f'{DOMAIN_SHORT}/']

# Movie-Related Constants
VALID_RATINGS = {0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5}
GENRES = [
    "action", "adventure", "animation", "comedy", "crime",
    "documentary", "drama", "family", "fantasy", "history",
    "horror", "music", "mystery", "romance", "science-fiction",
    "thriller", "tv-movie", "war", "western"
]

# Letterboxd Theme Colors
class Colors:
    BG = '#14181C'        # Letterboxd dark background
    BLUE = '#456'         # Letterboxd blue
    GREEN = '#00C030'     # Letterboxd green
    TEXT = '#9AB'         # Letterboxd text gray

# CSV Column Definitions
LIST_COLUMNS = ['Rank', 'Title', 'LetterboxdURI']
WATCHLIST_COLUMNS = ['Rank', 'Title', 'LetterboxdURI']

# Letterboxd official import format column mappings
# https://letterboxd.com/about/importing-data/
LETTERBOXD_COLUMNS = {
    'letterboxd_uri': 'LetterboxdURI',
    'tmdb_id': 'tmdbID',
    'imdb_id': 'imdbID',
    'title': 'Title',
    'year': 'Year',
    'directors': 'Directors',
    'rating': 'Rating',
    'rating10': 'Rating10',
    'watched_date': 'WatchedDate',
    'rewatch': 'Rewatch',
    'tags': 'Tags',
    'review': 'Review',
    'rank': 'Rank'
}

REQUIRED_COLUMNS = ['LetterboxdURI', 'tmdbID', 'imdbID', 'Title']
RATING_RANGE = (0.5, 5.0)
RATING10_RANGE = (1, 10)

CSV_FORMAT_COLUMNS = {
    "Letterboxd": {"Title": "Title", "LetterboxdURI": "LetterboxdURI"},
    "TMDB": {"Title": "Name", "LetterboxdURI": "Letterboxd URI"}
}



"""
Letterboxd Diary Poster Downloader

Downloads movie posters from user's diary entries.
- Extract poster URLs from diary entries
- Download and organize posters by year
- Automatic directory structure creation
- Skip existing files with size checking
"""

from curl_cffi import requests
import sys
import os
from bs4 import Tag

from letterboxdpy import user
from letterboxdpy.core.scraper import parse_url
from letterboxdpy.utils.utils_parser import extract_json_ld_script
from fastfingertips.terminal_utils import get_input, args_exists
from letterboxdpy.utils.utils_file import BinaryFile
from letterboxdpy.utils.utils_directory import Directory


class Settings:
    def __init__(self, foldering=True, size_check=False):
        self.foldering = foldering # Create folders for each day
        self.size_check = size_check  # Check if file size already exists

class App:
    def __init__(self, username):
        # Get the directory where this script is located
        self.script_dir = os.path.dirname(os.path.abspath(__file__))
        self.EXPORTS_DIR = os.path.join(self.script_dir, "exports")
        self.EXPORTS_USERS_DIR = os.path.join(self.EXPORTS_DIR, "users")

        self.username = username.lower()
        self.USER_FOLDER = os.path.join(self.EXPORTS_USERS_DIR, self.username)
        self.USER_POSTERS_DIR = os.path.join(self.USER_FOLDER, "posters")

        self.user = user.User(self.username)
        self.data = self.user.get_diary()
        self.config = Settings()

        self.foldering = self.config.foldering
        self.size_check = self.config.size_check

    def _extract_from_json_ld(self, page) -> str | None:
        data = extract_json_ld_script(page)
        if data and 'image' in data:
            return data['image']
        return None

    def _extract_from_meta(self, page) -> str | None:
        meta = page.find("meta", property="og:image")
        if isinstance(meta, Tag):
            content = meta.get('content')
            if isinstance(content, str):
                return content
        return None

    def _extract_from_dom(self, page) -> str | None:
        poster_div = page.find("div", class_="poster")
        if isinstance(poster_div, Tag):
            img = poster_div.find("img")
            if isinstance(img, Tag):
                src = img.get("src")
                if isinstance(src, str):
                    return src
        return None

    def _extract_from_ajax(self, slug: str) -> str | None:
        poster_ajax = f"https://letterboxd.com/ajax/poster/film/{slug}/std/500x750/"
        poster_page = parse_url(poster_ajax)
        img = poster_page.find('img')
        if isinstance(img, Tag):
            srcset = img.get('srcset')
            if isinstance(srcset, str):
                return srcset.split('?')[0]
        return None

    def get_poster_url(self, slug: str) -> str:
        # Method 1: Scrape film page
        try:
            film_url = f"https://letterboxd.com/film/{slug}/"
            page = parse_url(film_url)
            
            if url := self._extract_from_json_ld(page):
                return url
            if url := self._extract_from_meta(page):
                return url
            if url := self._extract_from_dom(page):
                return url
        except Exception:
            pass

        # Method 2: Legacy AJAX endpoint (fallback)
        try:
            if url := self._extract_from_ajax(slug):
                return url
        except Exception:
            pass

        raise LookupError(f"Poster not found: {slug}")

    def _download_poster(self, slug: str, file_path: str, count: int) -> None:
        """Download poster and save to file with size check."""
        poster_url = self.get_poster_url(slug)
        response = requests.get(poster_url)

        if os.path.exists(file_path):
            if int(os.stat(file_path).st_size) == int(response.headers['Content-Length']):
                print(f'{count} - File already exists and has same size as new file, skipping..')
                return
            print(f'Rewriting {file_path}..')

        BinaryFile.save(file_path, response.content)
        print(f'{count} - Wrote {file_path}')

    def _resolve_file_path(self, date: dict, slug: str, years_dir: str | None, previous_year: str | None) -> tuple[str, str | None]:
        """Calculate target file path and handle year directory creation."""
        file_date = "-".join(map(str, date.values()))
        filename = f"{file_date}_{slug}.jpg"
        
        updated_year = previous_year

        if self.foldering and years_dir:
            current_year = str(date['year'])
            year_dir = os.path.join(years_dir, current_year)
            
            if previous_year != current_year:
                Directory.check(year_dir)
                updated_year = current_year
                
            return os.path.join(year_dir, filename), updated_year
            
        return os.path.join(self.USER_POSTERS_DIR, filename), previous_year

    def _should_skip_download(self, file_path: str, count: int) -> bool:
        """Determines if the download should be skipped based on file existence."""
        if not os.path.exists(file_path):
            return False
            
        if not self.size_check:
            return True
            
        print(f'{count} - Poster file already exists, checking size..')
        return False

    def run(self):
        count = self.data['count']
        entries = self.data['entries']
        already_start = 0

        if not count:
            print('No entries found')
            return

        print(f'Processing {count} entries..')

        Directory.check(
            self.EXPORTS_DIR,
            self.EXPORTS_USERS_DIR,
            self.USER_FOLDER,
            self.USER_POSTERS_DIR
        )

        years_dir = None
        previous_year = None

        if self.foldering:
            years_dir = os.path.join(self.USER_POSTERS_DIR, 'years')
            Directory.check(years_dir)

        for v in entries.values():
            file_path, previous_year = self._resolve_file_path(
                v['date'], v['slug'], years_dir, previous_year
            )

            if self._should_skip_download(file_path, count):
                if not already_start:
                    already_start = count
                count -= 1
                continue

            if already_start and (already_start - count) > 1:
                print(f'Have already processed {already_start - count} entries, skipping {count}..')
                already_start = 0

            try:
                self._download_poster(v['slug'], file_path, count)
            except Exception as e:
                print(f"Error downloadable {v['slug']}: {e}")

            count -= 1

        print('Processing complete!')
        click_url = 'file:///' + self.USER_POSTERS_DIR.replace("\\", "/")
        print('At', click_url)


if __name__ == '__main__':
    if not args_exists():
        print(f'Quick usage: python {sys.argv[0]} <username>')

    username = get_input('Enter username: ', index=1)
    app = App(username)
    app.run()



"""
Letterboxd List Search and Export Tool

Searches for Letterboxd lists and exports them to CSV files.
- Search lists by query
- Export multiple lists to CSV format
- Automatic directory creation
- Batch processing support
"""

import sys
import os

from letterboxdpy.search import Search
from letterboxdpy.list import List
from fastfingertips.terminal_utils import get_input, args_exists
from letterboxdpy.utils.utils_file import CsvFile, build_path
from letterboxdpy.utils.utils_directory import Directory

def save_results_to_csv(list_instance: List, csv_file: str) -> None:
    """Saves movie list results to a CSV file."""
    # Get the directory where this script is located
    script_dir = os.path.dirname(os.path.abspath(__file__))
    directory = build_path(script_dir, 'exports', 'lists')
    Directory.create(directory, silent=True)

    filepath = build_path(directory, csv_file.replace('.csv', ''))
    movies = list_instance.movies
    
    rows = [[movie_data['url'], movie_data['name']] for movie_data in movies.values()]
    CsvFile.save(filepath, rows, headers=['LetterboxdURI', 'Title'])
    
    print(f"Data successfully saved to {csv_file}. Movies: {len(movies)}")

if __name__ == "__main__":
    if not args_exists():
        print(f'Quick usage: python {sys.argv[0]} <search_query> <max_lists>')

    search_query = get_input("Enter your search query for lists: ", index=1)

    search_instance = Search(search_query, "lists")
    search_data = search_instance.results

    if search_data['available']:
        results = search_data['results']
        search_count = search_data['count']

        print(f'Found {search_count} lists. ')
        max_lists = get_input('How many to export? (0 for all): ', index=2, expected_type=int)

        if max_lists == 0:
            max_lists = search_count

        print(f'Exporting first {max_lists} lists...')
        results = results[:max_lists]

        for result in results:
            list_slug = result['slug']
            list_owner_username = result['owner']['username']

            list_instance = List(list_owner_username, list_slug)
            csv_filename = f"{list_owner_username}_{list_slug}.csv"
            save_results_to_csv(list_instance, csv_filename)
    else:
        print(f'No lists found for "{search_query}".')



class Avatar:
    """Class to manage avatar URLs and upscale them if necessary."""

    # Default upscale size
    UPSCALE_SIZE = (1000, 1000)
    # List of default sizes to check against
    DEFAULT_SIZES = [(80, 80), (220, 220)]

    def __init__(self, url: str) -> None:
        """Initialize Avatar with the provided URL."""
        self.top_level = url.split('.')[0].split('//')[1]
        # Top levels: avatar:a, statics:s, secure
        self.avatar_exists = self.top_level == 'a'
        # Storing the URL without query parameters if the avatar exists
        self.url = url.split('?')[0] if self.avatar_exists else url
        # Initializing data dictionary with the initial state
        self.data = {
            'exists': self.avatar_exists,
            'upscaled': False,
            'url': self.url
        }
        # Storing a copy of data for internal use
        self._upscaled_data = self.data.copy()

    @property
    def upscaled_data(self) -> dict:
        """Return upscaled avatar data if applicable."""
        if self.avatar_exists:
            for default_size in self.DEFAULT_SIZES:
                pattern_default = '-0-'.join(map(str, default_size))
                # If a match is found, update the data with upscaled information
                if pattern_default in self.url:
                    pattern_upscale = '-0-'.join(map(str, self.UPSCALE_SIZE))
                    self._upscaled_data.update({
                        'upscaled': True,
                        'url': self.url.replace(pattern_default, pattern_upscale)
                    })
        return self._upscaled_data


if __name__ == '__main__':
    try:
        print(Avatar('https://unknown.example.com/test.png').upscaled_data)
        print(Avatar('https://s.example.com/a/0-220-0-220.png').upscaled_data)
        print(Avatar('https://a.example.com/a/0-220-0-220.png').upscaled_data)
        print(Avatar('https://a.example.com/a/0-80-0-80.png').upscaled_data)
    except Exception as e:
        raise RuntimeError(f"An error occurred: {e}")



"""
Letterboxd User Data Exporter

Exports comprehensive user data from Letterboxd profiles.
- Export all user data (films, reviews, lists, followers, etc.)
- Automatic JSON file generation
- Organized directory structure
- Progress tracking and timing
"""

import time
import os

from letterboxdpy import user
from letterboxdpy.utils.utils_string import strip_prefix
from fastfingertips.terminal_utils import get_input
from letterboxdpy.utils.utils_file import build_path, JsonFile, build_click_url
from letterboxdpy.utils.utils_directory import check_and_create_dirs

# -- MAIN --

username = get_input('Enter username: ', index=1)
user_instance = user.User(username)

# Get the directory where this script is located (examples folder)
script_dir = os.path.dirname(os.path.abspath(__file__))

# Export directories
EXPORTS_DIR = build_path(script_dir, 'exports')
USERS_FOLDER = build_path(EXPORTS_DIR, 'users')
USER_FOLDER = build_path(USERS_FOLDER, user_instance.username)
directories = [EXPORTS_DIR, USERS_FOLDER, USER_FOLDER]
check_and_create_dirs(directories)

start_time = time.time()

# Save user instance data
user_data_path = build_path(USER_FOLDER, 'user')
JsonFile.save(user_data_path, user_instance.jsonify())

# Export data for each method
# If you want to add a new method, add it here
# With arg: [user.User.user_watchlist, {'filters': {'genre': ['action', '-drama']}}],
methods = [
    user.User.get_activity,
    user.User.get_activity_following,
    user.User.get_diary,
    user.User.get_wrapped,
    user.User.get_films,
    [user.User.get_films_by_rating, {'rating':5}],
    user.User.get_films_not_rated,
    user.User.get_genre_info,
    user.User.get_liked_films,
    user.User.get_liked_reviews,
    user.User.get_lists,
    user.User.get_following,
    user.User.get_followers,
    user.User.get_reviews,
    user.User.get_user_tags,
    user.User.get_watchlist_movies,
    user.User.get_watchlist,
]
methods_str_length = len(str(len(methods)))

print('\nExporting data...')
for no, method in enumerate(methods, 1):
    method_start_time = time.time()

    args = {}
    if isinstance(method, list):
        method, args = method

    method_name = method.__name__
    method_name_without_prefix = strip_prefix(method_name)

    os.system(f'title [{len(methods)}/{no:0>{methods_str_length}}] Exporting {method_name}...')
    print(f'[{len(methods)}/{no:0>{methods_str_length}}]: Processing "{method_name}" method',
          end=f' with args: {args}...\r' if args else '...\r')

    data = method(user_instance, **args) if args else method(user_instance)

    file_path = build_path(USER_FOLDER, method_name_without_prefix)
    JsonFile.save(file_path, data)

    print(f'{time.time() - method_start_time:<7.2f} seconds - {method_name:<22} - {build_click_url(file_path)}.json')

os.system('title Completed!')
print('\nProcessing complete!')
print(f'\tTotal time: {time.time() - start_time:.2f} seconds')

print('\tAt', build_click_url(USER_FOLDER), end='\n\n')
os.system('pause')



if __loader__.name == '__main__':
    import sys
    sys.path.append(sys.path[0] + '/..')

from letterboxdpy.utils.utils_transform import get_ajax_url
from letterboxdpy.core.decorators import assert_instance
from letterboxdpy.core.scraper import parse_url
from letterboxdpy.utils.utils_url import get_page_url
from letterboxdpy.utils.movies_extractor import extract_movies_from_horizontal_list, extract_movies_from_vertical_list

class Films:
    """Fetch movies from Letterboxd based on different URLs."""
    VERTICAL_MAX = 20*5
    HORIZONTAL_MAX = 12*6

    def __init__(self, url: str):
        """Initialize Films class with a URL."""
        self.url = url
        self.ajax_url = get_ajax_url(url)
        self._movies = None

    @property
    def movies(self) -> dict:
        """Get movies from the URL."""
        if self._movies is None:
            self._movies = self.get_movies()
        return self._movies

    @property
    def count(self) -> int:
        """Return the count of movies."""
        return len(self.movies)

    # Magic Methods
    def __len__(self) -> int:
        return self.count

    def __getitem__(self, key: str):
        return self.movies[key]

    def get_movies(self) -> dict:
        """Scrape and return a dictionary of movies from Letterboxd."""
        page = 1
        movies = {}

        while True:
            page_url = get_page_url(self.ajax_url, page)
            dom = parse_url(page_url)

            if '.com/films/' in self.url:
                # https://letterboxd.com/films/popular/
                # https://letterboxd.com/films/like/v-for-vendetta/
                new_movies = extract_movies_from_horizontal_list(dom)
                movies |= new_movies
                if len(new_movies) < self.HORIZONTAL_MAX:
                    break
            elif '.com/film/' in self.url:
                # https://letterboxd.com/film/the-shawshank-redemption/similar/
                new_movies = extract_movies_from_vertical_list(dom)
                movies |= new_movies
                if len(new_movies) < self.VERTICAL_MAX:
                    break

            page += 1

        return movies

class Future:
    ARGS = ['name', 'release', 'release-earliest', 'rating',
          'rating-lowest', 'shortest', 'longest']

    def get_movies_with_args(args: list) -> dict:
        # by 
        pass

    def get_with_language(language: str):
        pass

    def get_with_country(country: str):
        pass

    def get_with_year(year: int):
        pass

    def get_with_actor(actor: str):
        pass

    def get_with_director(director: str):
        pass

    def get_with_writer(writer: str):
        pass

def get_upcoming_movies() -> dict:
    BASE_URL = "https://letterboxd.com/films/popular/this/week/upcoming/"
    return Films(BASE_URL).movies

@assert_instance(int)
def get_movies_by_decade(decade: int) -> dict:
    BASE_URL = f"https://letterboxd.com/films/ajax/popular/this/week/decade/{decade}s/"
    return Films(BASE_URL).movies

@assert_instance(int)
def get_movies_by_year(year: int) -> dict:
    BASE_URL = f"https://letterboxd.com/films/ajax/popular/this/week/year/{year}/"
    return Films(BASE_URL).movies

@assert_instance(str)
def get_movies_by_genre(genre: str) -> dict:
    """
    action, adventure, animation, comedy, crime, documentary,
    drama, family, fantasy, history, horror, music, mystery,
    romance, science-fiction, thriller, tv-movie, war, western
    """
    BASE_URL = f"https://letterboxd.com/films/ajax/genre/{genre}"
    return Films(BASE_URL).movies

@assert_instance(str)
def get_movies_by_service(service: str) -> dict:
    """
    netflix, hulu, prime-video, disney-plus, itv-play, apple-tv, 
    youtube-premium, amazon-prime-video, hbo-max, peacock, ...
    """
    BASE_URL = f"https://letterboxd.com/films/popular/this/week/on/{service}/"
    return Films(BASE_URL).movies

@assert_instance(str)
def get_movies_by_theme(theme: str) -> dict:
    BASE_URL = f"https://letterboxd.com/films/ajax/theme/{theme}"
    return Films(BASE_URL).movies

@assert_instance(str)
def get_movies_by_nanogenre(nanogenre: str) -> dict:
    BASE_URL = f"https://letterboxd.com/films/ajax/nanogenre/{nanogenre}/"
    return Films(BASE_URL).movies

@assert_instance(str)
def get_movies_by_mini_theme(theme: str) -> dict:
    BASE_URL = f"https://letterboxd.com/films/ajax/mini-theme/{theme}"
    return Films(BASE_URL).movies

def print_movies(movies, title=None, max_count=None):
    """Print movies in a formatted list."""
    if title:
        print(f"\n{title} -- ({len(movies)} movies)", end=f"\n{'*'*8*2*2}\n")
    for movie_no, (movie_id, movie) in enumerate(movies.items(), start=1):
        if max_count and movie_no > max_count:
            break
        print(f"{movie_no:<8} {movie_id:<8} {movie['name']}")
    print(f"{'*'*8*2*2}\n")

if __name__ == "__main__":
    sys.stdout.reconfigure(encoding='utf-8')

    # Movies similar to "V for Vendetta" are retrieved and printed.
    # https://letterboxd.com/films/like/v-for-vendetta/
    from letterboxdpy.movie import Movie
    movie_instance = Movie("v-for-vendetta")
    movies = movie_instance.get_similar_movies()
    print_movies(movies, title=f"Similar to {movie_instance.slug}")

    # Popular movies from the year 2027 are retrieved and displayed.
    # https://letterboxd.com/films/popular/this/week/year/2027/
    year = 2027
    movies = get_movies_by_year(year)
    print_movies(movies, title=f"Movies from {year}")



import re

from letterboxdpy.core.scraper import parse_url
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.utils.utils_parser import get_meta_content, get_movie_count_from_meta, get_body_content
from fastfingertips.url_utils import urls_match
from letterboxdpy.utils.movies_extractor import extract_movies_from_vertical_list
from letterboxdpy.utils.date_utils import DateUtils
from letterboxdpy.utils.utils_url import get_page_url


class ListMetaData(dict):
    """Type definition for list metadata"""
    pass


class UserList:
    LIST_PATTERN = f'{DOMAIN}/%s/list/%s'
    LIST_ITEMS_PER_PAGE = 12*5

    def __init__(self, username: str, slug: str) -> None:
        assert re.match("^[A-Za-z0-9_]+$", username), "Invalid author"

        self.username = username
        self.slug = slug
        self.url = self.LIST_PATTERN % (username, slug) 
        self.dom = parse_url(self.url)

    def __str__(self) -> str:
        return f"Not printable object of type: {self.__class__.__name__}"

    def get_title(self) -> str: return extract_title(self.dom)
    def get_author(self) -> str: return extract_author(self.dom)
    def get_description(self) -> str: return extract_description(self.dom)
    def get_date_created(self) -> list: return extract_date_created(self.dom)
    def get_date_updated(self) -> list: return extract_date_updated(self.dom)
    def get_tags(self) -> list: return extract_tags(self.dom)
    def get_movies(self) -> dict: return extract_movies(self.url, self.LIST_ITEMS_PER_PAGE)
    def get_count(self) -> int: return extract_count(self.dom)
    def get_list_id(self) -> str | None: return extract_list_id(self.dom)
    def get_list_meta(self, url: str) -> ListMetaData: return extract_list_meta(self.dom, url)

def extract_list_id(dom) -> str | None:
    """
    Extracts the list ID from the list page DOM.
    
    Args:
        dom: BeautifulSoup DOM object of the list page
        
    Returns:
        List ID as string or None if not found
    """
    try:
        # Method 1: Look for data-report-url attribute in report link
        report_link = dom.find('span', {'data-report-url': True})
        if report_link:
            report_url = report_link.get('data-report-url')
            if report_url and 'filmlist:' in report_url:
                # Extract ID from pattern like "/ajax/filmlist:30052453/report-form"
                import re
                match = re.search(r'filmlist:(\d+)', report_url)
                if match:
                    return match.group(1)
        
        # Method 2: Look for data-popmenu-id attribute
        report_menu = dom.find('a', {'data-popmenu-id': True})
        if report_menu:
            popmenu_id = report_menu.get('data-popmenu-id')
            if popmenu_id and 'list-' in popmenu_id:
                # Extract ID from pattern like "report-member-username-list-30052453"
                import re
                match = re.search(r'list-(\d+)$', popmenu_id)
                if match:
                    return match.group(1)
        
        return None
    except Exception as e:
        print(f"Error extracting list ID: {e}")
        return None

def extract_count(dom) -> int:
    """Extracts the number of films from the list DOM."""
    try:
        count = get_movie_count_from_meta(dom)
        if count is None:
            raise ValueError("Meta description not found or missing 'content' attribute.")
        return count
    except ValueError as e:
        raise RuntimeError("Failed to extract film count: " + str(e)) from e

def extract_movies(list_url: str, items_per_page) -> dict:
    data = {}

    page = 1
    while True:
        dom = parse_url(get_page_url(list_url, page))
        movies = extract_movies_from_vertical_list(dom)
        data |= movies

        if len(movies) < items_per_page:
            break

        page += 1

    return data

def extract_title(dom) -> str:
    return get_meta_content(dom, property='og:title')

def extract_author(dom) -> str:
    data = dom.find("span", attrs={'itemprop': 'name'})
    data = data.text if data else None
    return data

def extract_description(dom) -> str:
    return get_meta_content(dom, property='og:description')

def extract_date_created(dom) -> str | None:
    """Extract list creation date in ISO format."""
    # Look for published date span
    data = dom.find("span", {"class": "published is-updated"})
    if not data:
        data = dom.find("span", {"class": "published"})
    
    if data:
        time_element = data.findChild("time")
        if time_element and time_element.get('datetime'):
            return DateUtils.to_iso(time_element.get('datetime'))
    
    return None

def extract_date_updated(dom) -> str | None:
    """Extract list update date in ISO format."""
    # Look for updated date span
    data = dom.find("span", {"class": "updated"})
    if not data:
        # Use published date if no separate update date
        data = dom.find("span", {"class": "published"})
    
    if data:
        time_element = data.findChild("time")
        if time_element and time_element.get('datetime'):
            return DateUtils.to_iso(time_element.get('datetime'))
    
    return None

def extract_tags(dom) -> list:
    """
    Scraping the tag links from a Letterboxd list page and
    .. extracting just the tag names into a clean list.
    The decorator ensures a valid List instance is passed.
    """
    dom = dom.find("ul", {"class": ["tags"]})

    data = []

    if dom:
        dom = dom.findChildren("a")
        for item in dom:
            data.append(item.text)

    return data

def extract_list_meta(dom, url: str) -> ListMetaData:
    """
    Extracts metadata from a Letterboxd list page.
    Args:
        dom: BeautifulSoup DOM object
        url: The original URL of the list
    Returns:
        ListMetaData: A dictionary containing list metadata and status
    """
    data: ListMetaData = {
        'url': None,
        'title': None,
        'owner': None,
        'list_id': None,
        'is_available': False,
        'error': None
    }

    try:
        # Extract basic metadata
        list_url = get_meta_content(dom, property='og:url')
        list_title = get_meta_content(dom, property='og:title')
        list_owner = get_body_content(dom, 'data-owner')
        list_id = extract_list_id(dom)

        # Check for URL redirection
        # symmetric=False: checks if url == list_url or url + "/" == list_url
        if not urls_match(url, list_url, symmetric=False):
            print(f'Redirected to {list_url}')

        # Update metadata
        data.update({
            'url': list_url,
            'title': list_title,
            'owner': list_owner,
            'list_id': list_id,
            'is_available': True
        })

    except AttributeError as e:
        data['error'] = f"Missing required metadata: {str(e)}"
        print(f"Metadata extraction error: {e}")
    except Exception as e:
        data['error'] = f"Unexpected error: {str(e)}"
        print(f"Unexpected error while checking the list: {e}")

    return data




from letterboxdpy.movie import Movie
import unittest


class TestMovie(unittest.TestCase):

    def setUp(self):
        self.movie = Movie("v-for-vendetta")

    def test_get_not_exists_banner_movie(self):
        instance = Movie("avatar-4") # upcoming 2029
        data = instance.banner
        self.assertIsNone(data)

    def test_get_exists_banner_movie(self):
        data = self.movie.banner
        self.assertIsNotNone(data)

    def test_get_movie_title(self):
        data = self.movie.title
        self.assertEqual(data, "V for Vendetta")

    def test_get_movie_year(self):
        data = self.movie.year
        self.assertEqual(data, 2005)

    def test_movie_original_title_nullable(self):
        data = self.movie.original_title
        self.assertIsNone(data)
    
    def test_non_english_movie_original_title(self):
        movie = Movie("parasite-2019")
        self.assertEqual(movie.title, "Parasite")
        self.assertIsNotNone(movie.original_title)
        self.assertNotEqual(movie.title, movie.original_title)
        self.assertEqual(movie.original_title, "기생충")

    
if __name__ == '__main__':
    unittest.main()


"""
Shared list extraction utilities for Letterboxd list pages.
This module provides common functionality for extracting list data
from user lists, movie lists, and individual list pages.
"""

from letterboxdpy.utils.utils_parser import extract_and_convert_shorthand
from fastfingertips.string_utils import extract_number_from_text
from letterboxdpy.core.scraper import parse_url
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.utils.utils_url import extract_path_segment, get_page_url


class ListsExtractor:
    """Common lists collection extraction functionality.
    
    Extracts collections of lists from Letterboxd pages:
    - User's created lists
    - Lists containing a specific movie
    - Popular lists containing a movie
    """
    
    # Shared selectors for all list types
    SELECTORS = {
        'list_set': ('section', {'class': 'list-set'}),
        'lists': ('section', {'class': 'list'}),
        'title': ('h2', {'class': 'title'}),
        'description': ('div', {'class': 'body-text'}),
        'value': ('span', {'class': 'value'}),
        'likes': ('a', {'class': 'icon-like'}),
        'comments': ('a', {'class': 'icon-comment'}),
    }
    
    LISTS_PER_PAGE = 12
    
    @classmethod
    def from_url(cls, base_url: str, max_lists: int | None = None) -> dict:
        """
        Extract lists collection from URL.
        
        Args:
            base_url: Base URL without page parameter
            max_lists: Maximum number of lists to return (optional limit)
            
        Returns:
            dict: Contains 'lists', 'count', 'last_page'
        """
        data = {'limit': max_lists, 'count': 0, 'last_page': 1, 'lists': {}}
        page = 1
        
        while True:
            lists = cls._fetch_page_data(base_url, page)
            
            if not lists:
                break

            for item in lists:
                list_data = cls._extract_list_data(item)
                data['lists'] |= list_data

                if max_lists and len(data['lists']) >= max_lists:
                    # Limit reached
                    data['limit'] = True
                    break

            if data['limit'] or len(lists) < cls.LISTS_PER_PAGE:
                # Is last page or limit reached
                break

            page += 1

        data['count'] = len(data['lists'])
        data['last_page'] = page

        return data

    @classmethod
    def _fetch_page_data(cls, base_url: str, page: int):
        """Fetch and parse page data."""
        dom = parse_url(get_page_url(base_url, page))
        return dom.find_all('article', {'class': 'list-summary'})

    @classmethod
    def _extract_list_data(cls, item) -> dict:
        """Extract data from a list item."""

        def get_id() -> str:
            return item['data-film-list-id']

        def get_title() -> str:
            title_elem = item.find('h2', {'class': 'name'})
            return title_elem.text.strip()

        def get_description() -> str:
            description = item.find('div', {'class': ['notes', 'body-text']})
            if description:
                paragraphs = description.find_all('p')
                return '\n'.join([p.text for p in paragraphs])
            return ""

        def get_url() -> str:
            title_elem = item.find('h2', {'class': 'name'})
            return DOMAIN + title_elem.a['href']

        def get_slug() -> str | None:
            """
            extract list slug from url.
            example: 'https://letterboxd.com/user/list/my-list/' -> 'my-list'
            """
            return extract_path_segment(get_url(), after='/list/')

        def get_count() -> int:
            value_elem = item.find(*cls.SELECTORS['value'])
            if value_elem:
                # join=True: extracts and joins all digits (e.g. '1,234 films' -> 1234)
                count = extract_number_from_text(value_elem.text, join=True)
                return count if count is not None else 0
            return 0

        def get_likes() -> int:
            likes = item.find(*cls.SELECTORS['likes'])
            if likes:
                likes = extract_and_convert_shorthand(likes)
                return likes
            return 0

        def get_comments() -> int:
            comments = item.find(*cls.SELECTORS['comments'])
            if comments:
                return extract_and_convert_shorthand(comments)
            return 0

        return {
             get_id(): {
                'title': get_title(),
                'slug': get_slug(),
                'description': get_description(),
                'url': get_url(),
                'count': get_count(),
                'likes': get_likes(),
                'comments': get_comments()
                }
            }





"""
Letterboxd Ratings Histogram

Recreates the ratings distribution section of a Letterboxd profile with a clean, professional layout.
- Half-star tick labels (½, ★, ★½, …, ★★★★★)
- Shows username, total ratings, average, and most given rating
- Letterboxd-inspired color scheme
"""

import argparse
import sys
import math
import matplotlib.pyplot as plt
import numpy as np

from letterboxdpy.user import User
from fastfingertips.terminal_utils import get_input
from fastfingertips.string_utils import is_whitespace_or_empty
from letterboxdpy.constants.project import Colors


class LetterboxdRatingPlotter:
    def __init__(self, username=None):
        self.username = username

    def _get_star_labels(self):
        positions = np.arange(0.5, 5.5, 0.5)
        labels = []
        for r in positions:
            if math.isclose(r, 0.5):
                labels.append("½★")
            elif math.isclose(r % 1, 0.5):
                labels.append("★" * int(r) + "½")
            else:
                labels.append("★" * int(r))
        return positions, labels

    def _draw_header_and_stats(self, axis, stats_axis, total_count: int, average: float, most_given: float) -> None:
        if not self.username:
            return

        axis.text(0.02, 0.98, "R A T I N G S", transform=axis.transAxes,
                  fontsize=12, color=Colors.TEXT, weight='bold', va='top', family='monospace')
        axis.text(0.98, 0.98, f"{total_count:,}", transform=axis.transAxes,
                  fontsize=12, color=Colors.TEXT, weight='bold', va='top', ha='right')
        axis.text(0.02, 0.92, f"@{self.username}", transform=axis.transAxes,
                  fontsize=11, color='white', weight='bold', va='top')
        
        stats_text = f"Average: {average}★  •  Total: {total_count:,}  •  Most Given: {most_given}★"
        stats_axis.text(0.5, 0.5, stats_text,
                        ha='center', va='center', fontsize=11, color=Colors.TEXT, weight='bold')

    def _label_bars(self, axis, bars_, counts_) -> None:
        if len(counts_) == 0:
            return
        max_count = max(counts_)
        for bar, count in zip(bars_, counts_):
            if count == max_count:
                bar.set_color(Colors.GREEN)
                bar.set_alpha(1.0)
            if count > 0:
                axis.text(bar.get_x() + bar.get_width() / 2, count + max_count * 0.01,
                          str(int(count)), ha="center", va="bottom", fontsize=8,
                          color=Colors.TEXT, alpha=0.9)

    def _style_axes(self, axis, max_count) -> None:
        axis.set_xlim(0.25, 5.25)
        axis.set_ylim(0, max_count * 1.12 if max_count else 1)
        
        tick_positions, tick_labels = self._get_star_labels()
        axis.set_xticks(tick_positions)
        axis.set_xticklabels(tick_labels, fontsize=9, color=Colors.TEXT)
        axis.set_yticks([])
        
        for spine in axis.spines.values():
            spine.set_visible(False)
        axis.grid(True, axis='y', alpha=0.1, color=Colors.TEXT, linestyle='-')

    def create_plot(self, ratings: dict):
        """Create Letterboxd-style rating distribution plot with enhancements"""
        rating_positions = np.arange(0.5, 5.5, 0.5)
        rating_counts = np.array([ratings.get(rating, 0) for rating in rating_positions])
        total_ratings = int(rating_counts.sum())
        
        if total_ratings > 0:
            avg_calc = (rating_positions * rating_counts).sum() / total_ratings
            average_rating = round(float(avg_calc), 2)
            most_given_rating = float(rating_positions[rating_counts.argmax()])
        else:
            average_rating = 0.0
            most_given_rating = 0.0

        # Create Letterboxd-style plot
        fig, (ax, ax_stats) = plt.subplots(
            2, 1,
            figsize=(12, 8),
            gridspec_kw={"height_ratios": [0.86, 0.14], "hspace": 0},
            facecolor=Colors.BG,
        )
        
        # Safe window title setting
        if fig.canvas.manager:
            fig.canvas.manager.set_window_title(f"RATINGS - {self.username}")

        for a in (ax, ax_stats):
            a.set_facecolor(Colors.BG)
        ax_stats.axis('off')

        bars = ax.bar(rating_positions, rating_counts, width=0.45, color=Colors.BLUE, alpha=0.85)

        self._draw_header_and_stats(ax, ax_stats, total_ratings, average_rating, most_given_rating)
        self._label_bars(ax, bars, rating_counts)
        self._style_axes(ax, max(rating_counts) if len(rating_counts) else 0)

        plt.tight_layout()
        plt.show()

    def fetch_ratings(self, username=None) -> dict:
        """Fetch user ratings from Letterboxd"""
        target_user = username or self.username
        
        if not target_user:
             raise ValueError("Username must be provided")

        # Initialize with 0
        ratings = dict.fromkeys(np.arange(0.5, 5.5, 0.5), 0)
        
        print(f"Fetching ratings for @{target_user}...")
        movies = User(target_user).get_films()["movies"]
        print(f"Processing {len(movies)} rated movies...")
        
        for movie in movies.values():
            if rating := movie.get("rating"):
                try:
                    ratings[float(rating)] += 1
                except KeyError:
                    pass
        
        total_ratings = sum(ratings.values())
        print(f"Found {total_ratings} ratings. Creating plot...")
        return ratings

    def plot(self, username=None):
        """Fetch ratings and create plot"""
        if username:
            self.username = username

        try:
            ratings = self.fetch_ratings()
            if sum(ratings.values()) > 0:
                self.create_plot(ratings)
            else:
                print(f"No ratings found for user: {self.username}")
        except ValueError as e:
            print(f"Error: {e}")

    def run(self):
        """Main program loop"""
        try:
            sys.stdout.reconfigure(encoding="utf-8") # type: ignore
        except AttributeError:
            pass # Python < 3.7 or non-standard stdout

        parser = argparse.ArgumentParser(description="Visualize Letterboxd user rating distribution.")
        parser.add_argument("--user", help="Letterboxd username to analyze")

        args = parser.parse_args()
        
        username = args.user if not is_whitespace_or_empty(args.user) else None
        
        if not username:
            username = get_input("Enter Letterboxd username: ")
            
        self.plot(username)


def main():
    """Legacy function compatibility"""
    LetterboxdRatingPlotter().run()

if __name__ == "__main__":
    main()



if __name__ == '__main__':
    import sys
    sys.path.append(sys.path[0] + '/..')

from bs4 import BeautifulSoup, Tag
from curl_cffi import requests
from urllib.parse import quote
import time
import random

from letterboxdpy.utils.utils_file import JsonFile
from fastfingertips.terminal_utils import get_input
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.core.exceptions import (
    PageLoadError,
    InvalidResponseError,
    PrivateRouteError
)

class Scraper:
    """A class for scraping and parsing web pages."""

    _session = None
    headers = {
        "referer": DOMAIN,
        "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        "accept-encoding": "gzip, deflate, br",
        "accept-language": "en-US,en;q=0.9",
        "connection": "keep-alive"
    }
    builder = "lxml"
    timeout = (10, 30)  # (connect, read) in seconds; set None to disable

    def __init__(self, domain: str = headers['referer'], user_agent: str | None = None):
        """Initialize the scraper with the specified domain and user-agent."""
        self.headers = self.headers.copy()
        self.headers["referer"] = domain
        if user_agent:
            self.headers["user-agent"] = user_agent

    @classmethod
    def instance(cls):
        """Returns a singleton session instance."""
        if cls._session is None:
            cls._session = requests.Session()
        return cls._session

    @classmethod
    def get_page(cls, url: str) -> BeautifulSoup:
        """Fetch, check, and parse the HTML content from the specified URL."""
        response = cls._fetch(url)
        cls._check_for_errors(url, response)
        return cls._parse_html(response)

    @classmethod
    def _fetch(cls, url: str) -> requests.Response:
        """Fetch the HTML content from the specified URL using a session with retry logic."""
        max_retries = 3
        last_exception = None
        response = None
        
        for attempt in range(max_retries):
            try:
                session = cls.instance()
                response = session.get(url, headers=cls.headers, timeout=cls.timeout, impersonate="chrome")
                
                # If we get a 403, it might be a temporary block, try one more time after a short delay
                if response.status_code == 403 and attempt < max_retries - 1:
                    time.sleep(1 + random.random())
                    continue
                    
                return response
            except requests.errors.RequestsError as e:
                last_exception = e
                if attempt < max_retries - 1:
                    time.sleep(1 + attempt + random.random())
                    continue
                break
        
        if last_exception:
            raise PageLoadError(url, str(last_exception))
            
        if response is not None:
            return response
            
        raise PageLoadError(url, "Failed to fetch page after multiple attempts")

    @classmethod
    def _check_for_errors(cls, url: str, response: requests.Response) -> None:
        """Check the response for errors and raise an exception if found."""
        if response.status_code != 200:
            error_message = cls._get_error_message(response)
            formatted_error_message = cls._format_error(url, response, error_message)
            if response.status_code == 403:
                raise PrivateRouteError(formatted_error_message)
            raise InvalidResponseError(formatted_error_message)

    @classmethod
    def _get_error_message(cls, response: requests.Response) -> str:
        """Extract the error message from the response, if available."""
        dom = BeautifulSoup(response.text, cls.builder)
        message_section = dom.find("section", {"class": "message"})
        
        if isinstance(message_section, Tag):
            strong = message_section.find("strong")
            if strong:
                return strong.get_text()
                
        return "Unknown error occurred"

    @classmethod
    def _format_error(cls, url: str, response: requests.Response, message: str) -> str:
        """Format the error message for logging or raising exceptions."""
        return JsonFile.stringify({
            'code': response.status_code,
            'reason': str(response.reason),
            'url': url,
            'message': message
        }, indent=2)

    @classmethod
    def _parse_html(cls, response: requests.Response) -> BeautifulSoup:
        """Parse the HTML content from the response."""
        try:
            return BeautifulSoup(response.text, cls.builder)
        except Exception as e:
            raise InvalidResponseError(f"Error parsing response: {e}")

def parse_url(url: str) -> BeautifulSoup:
    """Fetch and parse the HTML content from the specified URL using the Scraper class."""
    return Scraper.get_page(url)

def url_encode(query: str, safe: str = '') -> str:
    """URL encode the given query."""
    return quote(query, safe=safe)

if __name__ == "__main__":
    sys.stdout.reconfigure(encoding='utf-8') # type: ignore

    input_domain = ''
    while not len(input_domain.strip()):
        input_domain = get_input('Enter url: ', index=0)

    print(f"Parsing {input_domain}...")

    parsed_dom = parse_url(input_domain)
    
    title_text = "No Title"
    if parsed_dom.title and parsed_dom.title.string:
        title_text = parsed_dom.title.string
        
    print(f"Title: {title_text}")

    input("Click Enter to see the DOM...")
    print(f"HTML: {parsed_dom.prettify()}")
    print("*" * 20 + "\nDone!")



from letterboxdpy.core.scraper import parse_url
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.utils.utils_parser import extract_json_ld_script, get_meta_content
from fastfingertips.string_utils import extract_number_from_text


class MovieProfile:
    """Movie profile page operations - main movie details."""
    
    def __init__(self, slug: str):
        """Initialize MovieProfile with a movie slug."""
        self.slug = slug
        self.url = f"{DOMAIN}/film/{slug}"
        self.dom = parse_url(self.url)
        
        # Get script data for some fields
        self.script = extract_json_ld_script(self.dom)
    
    # one line contents
    def get_id(self) -> str: return extract_movie_id(self.dom)
    def get_title(self) -> str: return extract_movie_title(self.dom)
    def get_original_title(self) -> str: return extract_movie_original_title(self.dom)
    def get_runtime(self) -> int: return extract_movie_runtime(self.dom)
    def get_rating(self) -> float: return extract_movie_rating(self.dom, self.script)
    def get_year(self) -> int: return extract_movie_year(self.dom, self.script)
    def get_tmdb_link(self) -> str: return extract_movie_tmdb_link(self.dom)
    def get_imdb_link(self) -> str: return extract_movie_imdb_link(self.dom)
    def get_poster(self) -> str: return extract_movie_poster(self.script)
    def get_banner(self) -> str: return extract_movie_banner(self.dom)
    def get_tagline(self) -> str: return extract_movie_tagline(self.dom)
    
    # long contents
    def get_description(self) -> str: return extract_movie_description(self.dom)
    def get_trailer(self) -> dict: return extract_movie_trailer(self.dom)
    def get_alternative_titles(self) -> list: return extract_movie_alternative_titles(self.dom)
    def get_details(self) -> list: return extract_movie_details(self.dom)
    def get_genres(self) -> list: return extract_movie_genres(self.dom, self.slug)
    def get_cast(self) -> list: return extract_movie_cast(self.dom)
    def get_crew(self) -> dict: return extract_movie_crew(self.dom)
    def get_popular_reviews(self) -> list: return extract_movie_popular_reviews(self.dom)

# ONE LINE CONTENTS

def extract_movie_id(dom):
    """Extract movie ID from DOM."""
    elem = dom.find('span', 'block-flag-wrapper')
    elem = elem.find('a')
    # join=True: extracts and joins all digits (e.g. data-report-url)
    return extract_number_from_text(elem.get('data-report-url'), join=True)

def extract_movie_title(dom):
    """Extract movie title from DOM."""
    h1_elem = dom.find("h1", {"class": ["primaryname"]})
    if h1_elem:
        span_elem = h1_elem.find("span", {"class": ["name"]})
        return span_elem.text if span_elem else h1_elem.text
    else:
        elem = dom.find("h1", {"class": ["filmtitle"]})
        return elem.text if elem else None

def extract_movie_original_title(dom):
    """Extract movie original title from DOM."""
    elem = dom.find("h2", {"class": ["originalname"]})
    return elem.text.strip() if elem else None

def extract_movie_runtime(dom):
    """Extract movie runtime from DOM."""
    elem = dom.find("p", {"class": ["text-footer"]})
    # join=True: extracts and joins all digits (e.g. '124 mins' -> 124)
    return extract_number_from_text(elem.text, join=True) if elem else None

def extract_movie_rating(dom, script=None):
    """Extract movie rating from DOM."""
    elem = dom.find('span', attrs={'class': 'average-rating'})
    rating = elem.text if elem else None
    try:
        rating = rating if rating else (
            script['aggregateRating']['ratingValue'] if script else None
        )
        return float(rating) if rating else None
    except (KeyError, ValueError):
        return None

def extract_movie_year(dom, script=None):
    """Extract movie year from DOM."""
    elem = dom.find('div', {'class': 'releaseyear'})
    year = elem.text if elem else None
    try:
        year = year if year else (
            script['releasedEvent'][0]['startDate'] if script else None
        )
        return int(year) if year else None
    except (KeyError, ValueError):
        return None

def extract_movie_tmdb_link(dom):
    """Extract TMDB link from DOM."""
    a = dom.find("a", {"data-track-action": ["TMDB"]})
    return a['href'] if a else None

def extract_movie_imdb_link(dom):
    """Extract IMDB link from DOM."""
    a = dom.find("a", {"data-track-action": ["IMDb"]})
    return a['href'] if a else None

def extract_movie_poster(script):
    """Extract movie poster from script."""
    # crop: list=(1500, 1000)
    # .replace('230-0-345', f'{crop[0]}-0-{crop[1]}')
    if script:
        poster = script['image'] if 'image' in script else None
        return poster.split('?')[0] if poster else None
    else:
        return None

def extract_movie_banner(dom):
    """Extract movie banner from DOM."""
    elem = dom.find("div", {"id": ["backdrop"]})
    if not elem:
        return None
    
    # Try high-res first, fallback to standard
    if "data-backdrop2x" in elem.attrs:
        return elem['data-backdrop2x'].split('?')[0]
    elif "data-backdrop" in elem.attrs:
        return elem['data-backdrop'].split('?')[0]
    
    return None

def extract_movie_tagline(dom):
    """Extract movie tagline from DOM."""
    elem = dom.find(attrs={'class':'tagline'})
    return elem.text if elem else None

# LONG CONTENTS

def extract_movie_description(dom):
    """Extract movie description from DOM."""
    # elem_section = page.find("div", attrs={'class': 'truncate'}).text
    return get_meta_content(dom, name='description')

def extract_movie_popular_reviews(dom) -> list:
    """Extract popular reviews from main movie page."""
    container_section = dom.find("section", {"class": ["film-reviews"]})

    def get_text_or_none(element):
        return element.text.strip() if (element and element.text) else None

    def extract_reviewer_username(article):
        return article.get("data-person")

    def extract_reviewer_display_name(article):
        return get_text_or_none(article.find("strong", {"class": ["displayname"]}))

    def extract_review_link(article):
        context_link = article.find("a", {"class": ["context"]})
        href = context_link.get("href") if context_link else None
        return (DOMAIN + href) if href else None

    def extract_rating(article) -> float | None:
        """Extracts the rating element from a review and converts it to a numerical value (float)."""

        rating_span = article.find("span", class_="rating")

        # [LEGACY NOTE]: Original robust finding logic (kept as reference)
        # rating_span = article.find("span", {"class": ["rating"]})
        # if not (rating_span and rating_span.text):
        #     for span in article.find_all("span"):
        #         classes = span.get("class") or []
        #         if any((cls == "rating") or cls.startswith("rating") for cls in classes):
        #             rating_span = span
        #             break
        
        if rating_span:
            for cls in rating_span.get('class', []):
                if cls.startswith('rated-'):
                    try:
                        return int(cls.split('-')[-1]) / 2.0
                    except (ValueError, IndexError):
                        continue
        return None

    def extract_review_text(article):
        body_div = article.find("div", {"class": ["body-text"]})
        paragraph = body_div.find("p") if body_div else None
        return get_text_or_none(paragraph)

    reviews = []
    if container_section:
        article_elements = container_section.find_all("article", {"class": ["production-viewing"]})
        for article in article_elements:
            reviews.append({
                "user": {
                    "username": extract_reviewer_username(article),
                    "display_name": extract_reviewer_display_name(article)
                },  
                "link": extract_review_link(article),
                "rating": extract_rating(article),
                "review": extract_review_text(article),
            })

    return reviews

def extract_movie_trailer(dom):
    """Extract movie trailer from DOM."""
    elem = dom.find("p", {"class": ["trailer-link"]})

    if elem:
        elem = elem.a['href'].lstrip('/').split('?')[0]
        trailer_id = elem.split('/')[-1]

        return {
            'id': trailer_id,
            'link': f"https://www.youtube.com/watch?v={trailer_id}",
            'embed_url': f"https://{elem}",
        }
    return None

def extract_movie_alternative_titles(dom):
    """Extract movie alternative titles from DOM."""
    data = dom.find(attrs={'class': 'text-indentedlist'})
    text = data.text if data else None
    return [i.strip() for i in text.split(', ')] if text else None

def extract_movie_details(dom):
    """Extract movie details from DOM."""
    data = dom.find("div", {"id": ["tab-details"]})
    data = data.find_all("a") if data else []

    details = []
    for item in data:
        url_parts = item['href'].split('/')

        _ = url_parts[1] == 'films'
        item_type, slug = url_parts[1+_:3+_] 

        details.append({
            'type': item_type,
            'name': item.text.strip(),
            'slug': slug,
            'url': DOMAIN + "/".join(url_parts)
        })

    return details

def extract_movie_genres(dom, slug):
    """Extract movie genres from DOM."""
    data = dom.find(attrs={"id": ["tab-genres"]})
    data = data.find_all("a") if data else []

    genres = []
    for item in data:
        url_parts = item['href'].split('/')
        
        genres.append({
            'type': url_parts[2],
            'name': item.text,
            'slug': url_parts[3],
            'url': DOMAIN + "/".join(url_parts)
        })

    if genres and slug == genres[-1]['type']:
        genres.pop() # for show all button

    return genres


def extract_movie_cast(dom):
    """Extract movie cast from DOM."""
    data = dom.find("div", {"id": ["tab-cast"]})
    data = data.find_all("a", {"class": {"tooltip"}}) if data else []

    cast = []
    for person in data:
        name = person.text
        role_name = person['title'] if 'title' in person.attrs else None
        url = person['href']
        slug = url.split('/')[-2]
        
        cast.append({
            'name': name,
            'role_name': role_name,
            'slug': slug,
            'url': DOMAIN + url
        })

    return cast

def extract_movie_crew(dom):
    """Extract movie crew from DOM."""
    data = dom.find("div", {"id": ["tab-crew"]})
    data = data.find_all("a") if data else []

    crew = {}
    for person in data:
        name = person.text
        url = person['href']

        job, slug = filter(None, url.split('/'))
        job = job.replace('-', '_')

        if job not in crew:
            crew[job] = []

        crew[job].append({
            'name': name,
            'slug': slug,
            'url': DOMAIN + url
        })

    return crew



import os
from json import (
    dump as json_dump,
    load as json_load,
    loads as json_loads,
    dumps as json_dumps
)


class File:
    """Base utility class for file operations."""
    EXTENSION = ''
    
    @classmethod
    def _get_path(cls, path: str) -> str:
        """Get full path with extension."""
        if cls.EXTENSION and not path.endswith(cls.EXTENSION):
            return f'{path}{cls.EXTENSION}'
        return path
    
    @classmethod
    def exists(cls, path: str) -> bool:
        """Check if file exists."""
        return os.path.exists(cls._get_path(path))
    
    @classmethod
    def delete(cls, path: str) -> bool:
        """Delete file if exists. Returns True if deleted."""
        filepath = cls._get_path(path)
        if os.path.exists(filepath):
            os.remove(filepath)
            return True
        return False
    
    @staticmethod
    def save(path: str, data: dict, format: str = 'json') -> None:
        """Save data to file in the specified format."""
        if format == 'json':
            JsonFile.save(path, data)
        else:
            raise ValueError(f"Unsupported format '{format}'. Only 'json' is currently supported.")


class JsonFile(File):
    """Utility class for JSON file operations."""
    EXTENSION = '.json'
    
    @classmethod
    def save(cls, path: str, data: dict, indent: int = 2) -> None:
        """Save data to a JSON file."""
        with open(cls._get_path(path), 'w') as f:
            json_dump(data, f, indent=indent)
    
    @classmethod
    def load(cls, path: str) -> dict | None:
        """Load data from a JSON file. Returns None if file doesn't exist."""
        filepath = cls._get_path(path)
        if os.path.exists(filepath):
            with open(filepath, 'r', encoding='utf-8') as f:
                return json_load(f)
        return None
    
    @staticmethod
    def parse(text: str) -> dict | None:
        """Parse JSON from string. Returns None if parsing fails."""
        try:
            return json_loads(text)
        except (ValueError, TypeError):
            return None
    
    @staticmethod
    def stringify(data, indent: int = None, encoder=None, **kwargs) -> str:
        """Convert dict to JSON string. Supports custom encoder and extra args."""
        return json_dumps(data, indent=indent, cls=encoder, **kwargs)


class CsvFile(File):
    """Utility class for CSV file operations."""
    EXTENSION = '.csv'
    
    @classmethod
    def save(cls, path: str, rows: list, headers: list = None) -> None:
        """Save rows to a CSV file. First row can be headers."""
        import csv
        with open(cls._get_path(path), 'w', newline='', encoding='utf-8') as f:
            writer = csv.writer(f)
            if headers:
                writer.writerow(headers)
            writer.writerows(rows)
    
    @classmethod
    def load(cls, path: str) -> list | None:
        """Load rows from a CSV file. Returns None if file doesn't exist."""
        import csv
        filepath = cls._get_path(path)
        if os.path.exists(filepath):
            with open(filepath, 'r', encoding='utf-8') as f:
                return list(csv.reader(f))
        return None


class BinaryFile(File):
    """Utility class for binary file operations (images, etc)."""
    EXTENSION = ''
    
    @classmethod
    def save(cls, path: str, data: bytes) -> None:
        """Save binary data to file."""
        with open(cls._get_path(path), 'wb') as f:
            f.write(data)
    
    @classmethod
    def load(cls, path: str) -> bytes | None:
        """Load binary data from file. Returns None if file doesn't exist."""
        filepath = cls._get_path(path)
        if os.path.exists(filepath):
            with open(filepath, 'rb') as f:
                return f.read()
        return None


def build_path(*segments: str, normalize: bool = True) -> str:
    """Build and format file paths from the given segments."""
    path = os.path.join(*segments)
    if normalize:
        return os.path.normpath(path)
    return path

def build_click_url(file_path: str, protocol: str = 'file') -> str:
    """Build a clickable file URL with the specified protocol."""
    if protocol == 'file':
        return f"file:///{build_path(os.getcwd(), file_path).replace(os.sep, '/')}"
    elif protocol in ['http', 'https']:
        return f"{protocol}://{file_path}"
    else:
        raise ValueError(f"Unsupported protocol '{protocol}'")



"""
Letterboxd User Statistics Plotter

Creates visualizations of user movie watching patterns over time.
- Monthly and daily viewing statistics by year
- Multi-year comparison charts
- Letterboxd-inspired styling
"""

import matplotlib.pyplot as plt
from letterboxdpy.user import User
from letterboxdpy.constants.project import Colors, DAY_ABBREVIATIONS, MONTH_ABBREVIATIONS
import argparse
import sys
from datetime import datetime


class LetterboxdStatisticsPlotter:
    """Class for plotting Letterboxd user statistics."""
    
    def __init__(self, username: str):
        self.username = username
        self.stats_by_year = {}
        
    def gather_statistics_by_year(self, start_year: int, end_year: int) -> dict:
        """Fetch user statistics for each year."""
        self.stats_by_year = {}
        year_count = end_year - start_year + 1
        
        print(f"Fetching statistics for @{self.username}...")
        print(f"Processing {year_count} year(s): {start_year}-{end_year}")
        
        for year in range(start_year, end_year + 1):
            try:
                print(f"Fetching data for {year}...", end=" ")
                user = User(self.username)
                stats = user.get_wrapped(year)
                self.stats_by_year[year] = {
                    "monthly": stats.get("months"),
                    "daily": stats.get("days")
                }
                print("✓")
            except Exception:
                print("✗ (using empty data)")
                self.stats_by_year[year] = {
                    "monthly": dict.fromkeys(range(1, 13), 0),  # 12 months with 0
                    "daily": dict.fromkeys(range(1, 8), 0)      # 7 days with 0
                }
        
        print("Data collection complete. Creating plot...")
        return self.stats_by_year
    
    def plot_statistics(self) -> None:
        if not self.stats_by_year:
            return
        
        num_years = len(self.stats_by_year)
        
        # squeeze=False ensures axes is always a 2D array, simplifying logic
        fig, axes = plt.subplots(
            num_years, 2, 
            figsize=(12, 3 * num_years), 
            facecolor=Colors.BG, 
            squeeze=False
        )

        years_range = f"{min(self.stats_by_year.keys())}-{max(self.stats_by_year.keys())}"
        
        # Safe window title access
        if fig.canvas.manager:
            fig.canvas.manager.set_window_title(f'Letterboxd Statistics - {self.username} ({years_range})')
            
        fig.suptitle(f'{self.username} - Movies Watched ({years_range})', fontsize=16, color='white')

        def style_axes(ax):
            ax.set_facecolor(Colors.BG)
            ax.tick_params(colors=Colors.TEXT)
            for spine in ax.spines.values():
                spine.set_color(Colors.TEXT)
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['left'].set_color(Colors.TEXT)
            ax.spines['bottom'].set_color(Colors.TEXT)

        days_labels = DAY_ABBREVIATIONS
        months_labels = MONTH_ABBREVIATIONS
        
        for i, (year, stats) in enumerate(self.stats_by_year.items()):
            daily_data = stats.get('daily', {})
            monthly_data = stats.get('monthly', {})
            
            daily_values = [daily_data.get(day, 0) for day in range(1, 8)]
            monthly_values = [monthly_data.get(month, 0) for month in range(1, 13)]

            # Access subplots safely from 2D array
            ax_daily = axes[i, 0]
            ax_monthly = axes[i, 1]

            for ax in [ax_daily, ax_monthly]:
                style_axes(ax)

            ax_daily.bar(days_labels, daily_values, color=Colors.BLUE, alpha=0.85)
            ax_daily.set_title(f'{year} - Daily', color='white')
            ax_daily.set_ylabel('Movies', color=Colors.TEXT)

            ax_monthly.bar(months_labels, monthly_values, color=Colors.GREEN, alpha=0.85)
            ax_monthly.set_title(f'{year} - Monthly', color='white')
            ax_monthly.set_ylabel('Movies', color=Colors.TEXT)

        plt.tight_layout(rect=(0, 0, 1, 0.95))
        plt.show()

    def plot(self, start_year=None, end_year=None):
        """Gather statistics and create plot"""
        if start_year is None:
            current_year = datetime.now().year
            start_year = current_year - 1
        if end_year is None:
            end_year = datetime.now().year
            
        self.gather_statistics_by_year(start_year, end_year)
        
        if self.stats_by_year:
            self.plot_statistics()
        else:
            print(f"No statistics found for user: {self.username}")

    def run(self):
        """Main program loop"""
        try:
             sys.stdout.reconfigure(encoding="utf-8") # type: ignore
        except AttributeError:
             pass

        parser = argparse.ArgumentParser(description="Visualize Letterboxd user statistics")
        parser.add_argument("--user", help="Letterboxd username")
        current_year = datetime.now().year
        parser.add_argument("--start-year", type=int, default=current_year-1, help=f"Start year (default: {current_year-1})")
        parser.add_argument("--end-year", type=int, default=current_year, help=f"End year (default: {current_year})")
        args = parser.parse_args()

        username = args.user
        if not username or not username.strip():
            username = input("Enter a Letterboxd username: ").strip()
            
        self.username = username
        self.plot(args.start_year, args.end_year)


def main():
    """Legacy function compatibility"""
    LetterboxdStatisticsPlotter("").run()


if __name__ == "__main__":
    main()



from letterboxdpy.search import Search, get_film_slug_from_title
import unittest


class TestSearch(unittest.TestCase):

    def setUp(self):
        self.movie_name = "V for Vendetta"
        self.q = Search(self.movie_name, 'films')

    def test_film_search(self):
        data = self.q.get_results()
        self.assertGreater(len(data['results']), 0)

    def test_film_search_with_max(self):
        data = self.q.get_results(max=1)
        self.assertEqual(data['count'], 1)
        self.assertEqual(len(data['results']), 1)

    def test_get_film_slug_from_title(self):
        slug = get_film_slug_from_title(self.movie_name)
        self.assertEqual(slug, 'v-for-vendetta')

if __name__ == '__main__':
    unittest.main()


from letterboxdpy.core.scraper import parse_url
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.core.exceptions import PageFetchError
from letterboxdpy.avatar import Avatar
from letterboxdpy.utils.utils_url import get_page_url


class UserNetwork:

    def __init__(self, username: str) -> None:
        self.username = username
        self.following_url = f"{DOMAIN}/{self.username}/following"
        self.followers_url = f"{DOMAIN}/{self.username}/followers"

    def get_following(self, page: int = 1, limit: int = None) -> dict: return extract_network(self.username, 'following', page=page, limit=limit)
    def get_followers(self, page: int = 1, limit: int = None) -> dict: return extract_network(self.username, 'followers', page=page, limit=limit)

def extract_network(username: str, section: str, limit: int = None, page: int = 1) -> dict:
    """
    Fetches the specified network section ('followers' or 'following') for the user.
    limit: Optional maximum number of pages to fetch.
    page: Optional starting page number.
    """
    assert section in ['followers', 'following'], "Section must be either 'followers' or 'following'"

    BASE_URL = f"{DOMAIN}/{username}/{section}"
    PERSONS_PER_PAGE = 25

    def fetch_page(page_num: int):
        """Fetches a single page of the user's network section."""
        try:
            return parse_url(get_page_url(BASE_URL, page_num))
        except Exception as e:
            raise PageFetchError(f"Failed to fetch page {page_num}: {e}") from e

    def extract_persons(dom) -> dict:
        """Extracts persons from a DOM object and returns them as a dictionary."""
        persons_dict = {}
        
        # Find the member table
        member_table = dom.find('table', class_='member-table')
        if not member_table:
            return persons_dict
            
        # Find all user rows
        user_rows = member_table.find_all('tr')
        
        for row in user_rows:
            # Get the person summary div
            person_summary = row.find('div', class_='person-summary')
            if not person_summary:
                continue
                
            # Extract avatar info
            avatar_link = person_summary.find('a', class_='avatar')
            if not avatar_link:
                continue
                
            # Extract basic info
            username = avatar_link['href'].replace('/', '')
            avatar_img = avatar_link.find('img')
            display_name = avatar_img['alt'] if avatar_img else username
            avatar_url = avatar_img['src'] if avatar_img else ''
            
            # Process avatar with Avatar class
            avatar_data = Avatar(avatar_url).upscaled_data if avatar_url else {'exists': False, 'upscaled': False, 'url': ''}
            
            # Extract name link
            name_link = person_summary.find('a', class_='name')
            if name_link:
                display_name = name_link.get_text(strip=True)
            
            # Extract metadata (followers, following)
            metadata = person_summary.find('small', class_='metadata')
            followers_count = None
            following_count = None
            
            if metadata:
                followers_link = metadata.find('a', href=lambda x: x and 'followers' in x)
                if followers_link:
                    followers_text = followers_link.get_text(strip=True)
                    # Extract number from "5 followers"
                    import re
                    followers_match = re.search(r'(\d+)', followers_text)
                    if followers_match:
                        followers_count = int(followers_match.group(1))
                
                following_link = metadata.find('a', href=lambda x: x and 'following' in x)
                if following_link:
                    following_text = following_link.get_text(strip=True)
                    # Extract number from "following 6"
                    following_match = re.search(r'(\d+)', following_text)
                    if following_match:
                        following_count = int(following_match.group(1))
            
            # Extract stats from other columns
            watched_cell = row.find('td', class_='col-watched')
            watched_count = None
            if watched_cell:
                watched_link = watched_cell.find('a')
                if watched_link:
                    watched_text = watched_link.get_text(strip=True)
                    import re
                    watched_match = re.search(r'(\d+)', watched_text)
                    if watched_match:
                        watched_count = int(watched_match.group(1))
            
            lists_cell = row.find('td', class_='col-lists')
            lists_count = None
            if lists_cell:
                lists_link = lists_cell.find('a')
                if lists_link:
                    lists_text = lists_link.get_text(strip=True)
                    import re
                    lists_match = re.search(r'(\d+)', lists_text)
                    if lists_match:
                        lists_count = int(lists_match.group(1))
            
            likes_cell = row.find('td', class_='col-likes')
            likes_count = None
            if likes_cell:
                likes_link = likes_cell.find('a')
                if likes_link:
                    likes_text = likes_link.get_text(strip=True)
                    import re
                    likes_match = re.search(r'(\d+)', likes_text)
                    if likes_match:
                        likes_count = int(likes_match.group(1))
            
            persons_dict[username] = {
                'username': username,
                'name': display_name,
                'url': f"{DOMAIN}/{username}",
                'avatar': avatar_data,
                'followers': followers_count,
                'following': following_count,
                'watched': watched_count,
                'lists': lists_count,
                'likes': likes_count
            }

        return persons_dict

    users_list = {}
    page_num = page
    fetched_count = 0

    while True:
        if limit is not None and fetched_count >= limit:
            break

        dom = fetch_page(page_num)
        persons = extract_persons(dom)
        users_list.update(persons)

        fetched_count += 1

        # Break if the number of persons fetched is less than a full page (end of list)
        if len(persons) < PERSONS_PER_PAGE :
            break

        page_num += 1

    return users_list



from letterboxdpy.constants.project import DOMAIN


class MovieSimilar:
    """Movie similar page operations - similar movies functionality."""
    
    def __init__(self, slug: str):
        """Initialize MovieSimilar with a movie slug."""
        self.slug = slug
        self.url = f"https://letterboxd.com/films/ajax/like/{slug}"
    
    def get_similar_movies(self) -> dict:
        """Get movies similar to this movie."""
        return extract_similar_movies(self.url)


def extract_similar_movies(url: str) -> dict:
    """Extract movies similar to the given movie."""
    from letterboxdpy.films import Films  # Avoid circular import
    
    # Using the AJAX endpoint for similar movies
    return Films(url).movies

if __name__ == "__main__":
    similar_instance = MovieSimilar("v-for-vendetta")

    print(f"Movie: {similar_instance.slug}")
    for id, data in similar_instance.get_similar_movies().items():
        print(id, data)



if __loader__.name == '__main__':
    import sys
    sys.path.append(sys.path[0] + '/..')

import os
import inspect
from letterboxdpy import user, movie, films, members, search, list

def get_defined_functions(module):
    """Returns a list of function names defined in the given module."""
    functions = []
    for name, obj in inspect.getmembers(module):
        if inspect.isfunction(obj) and inspect.getmodule(obj) == module:
            functions.append(name)
    return functions

def get_existing_md_files(directory):
    """Returns a list of .md files in the given directory without extension."""
    md_files = [f[:-3] for f in os.listdir(directory) if f.endswith('.md')]
    return md_files

def check_missing_md_files(functions, md_files):
    """Compares functions and .md files, returning functions without corresponding .md files."""
    missing_md = [func for func in functions if func not in md_files]
    return missing_md

def create_md_file_for_missing_function(func_name, module, directory):
    """Creates a .md file for a missing function with its signature."""
    file_path = os.path.join(directory, f"{func_name}.md")
    func = getattr(module, func_name)

    signature = str(inspect.signature(func))
    docstring = inspect.getdoc(func) or "No documentation provided."
    
    with open(file_path, 'w') as file:
        file.write(f'<h2 id="{func_name}">{func_name}{signature}</h2>\n\n')
        file.write(f'**Documentation:**\n\n{docstring}\n\n')
        file.write(f'[To be documented.](https://github.com/search?q=repo:nmcassa/letterboxdpy+{func_name})\n')

def check_modules_for_missing_md(modules):
    """Checks each module for missing .md files and prints the results."""
    base_directory = "."
    for module_name, module in modules.items():
        print(f"{module_name}:")
        function_names = get_defined_functions(module)
        md_directory = os.path.join(base_directory, module_name, 'funcs')
        
        if not os.path.exists(md_directory):
            print(f"Directory {md_directory} does not exist. Creating...")
            os.makedirs(md_directory, exist_ok=True)
        
        md_files = get_existing_md_files(md_directory)
        missing_md_files = check_missing_md_files(function_names, md_files)
        
        for func in missing_md_files:
            create_md_file_for_missing_function(func, module, md_directory)
            print(f"✗ {func}.md missing and created.")
        
        for func in function_names:
            if func in md_files:
                print(f"✓ {func}.md exists")
        
        if not missing_md_files:
            print("All functions have corresponding .md files.")
        print()

if __name__ == "__main__":
    modules = {
        'user': user,
        'movie': movie,
        'films': films,
        'members': members,
        'search': search,
        'list': list
    }
    check_modules_for_missing_md(modules)



import os


class Directory:
    """Utility class for directory operations."""
    
    @staticmethod
    def exists(path: str) -> bool:
        """Check if directory exists."""
        return os.path.isdir(path)
    
    @staticmethod
    def create(path: str, silent: bool = False) -> bool:
        """Create directory if it doesn't exist. Returns True if created."""
        try:
            if not os.path.exists(path):
                if not silent:
                    print(f'Directory created: {path}')
                os.makedirs(path, exist_ok=True)
                return True
            
            if not silent:
                print(f'Directory found: {path}')
            return False
        except OSError as e:
            if not silent:
                print(f"Error creating {path}: {e}")
            return False
    
    @staticmethod
    def delete(path: str) -> bool:
        """Delete empty directory. Returns True if deleted."""
        if os.path.isdir(path):
            os.rmdir(path)
            return True
        return False
    
    @staticmethod
    def list(path: str) -> list:
        """List contents of directory."""
        if os.path.isdir(path):
            return os.listdir(path)
        return []
    
    @staticmethod
    def check(*paths: str) -> None:
        """Check and create multiple directories. Prints status for each."""
        for path in paths:
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
                print(f'Directory created: {path}')
            else:
                print(f'Directory found: {path}')


def check_and_create_dirs(directories: list | str) -> None:
    """Checks if directories exist, creates them if not."""
    if isinstance(directories, str):
        directories = [directories]

    print('\nChecking directories...')
    for directory in directories:
        Directory.create(directory)
    print('All directories checked, continuing...', end='\n\n')



from letterboxdpy.core.scraper import parse_url
from letterboxdpy.utils.utils_parser import parse_review_date, parse_review_text
from letterboxdpy.constants.project import DOMAIN
from letterboxdpy.utils.utils_url import get_page_url


class UserReviews:

    def __init__(self, username: str) -> None:
        self.username = username
        self.url = f"{DOMAIN}/{self.username}/films/reviews"
        
    def get_reviews(self): return extract_user_reviews(self.url)

def extract_user_reviews(url: str) -> dict:
    '''
    Returns a dictionary containing user reviews. The keys are unique log IDs,
    and each value is a dictionary with details about the review,
    including movie information, review type, rating, review content, date, etc.
    '''
    LOGS_PER_PAGE = 12

    page = 0
    data = {'reviews': {}}
    while True:
        page += 1
        dom = parse_url(get_page_url(url, page))

        container = dom.find("div", {"class": ["viewing-list"]})

        if not container:
            # No container (div.viewing-list) found in the page.
            ...

        logs = container.find_all("article")

        if not logs:
            # No item (article) found in container.
            ...

        for log in logs:
            # Handle react structure
            react_component = log.parent.find("div", {"class": "react-component"}) or log.parent.div
            
            movie_name = log.a.text
            slug = react_component.get('data-item-slug') or react_component.get('data-film-slug')
            movie_id = react_component['data-film-id']
            # str   ^^^--- movie_id: unique id of the movie.
            # Find release year in spans
            release = None
            spans = log.find_all('span')
            for span in spans:
                if span.text and span.text.strip().isdigit() and len(span.text.strip()) == 4:
                    release = int(span.text.strip())
                    break
            movie_link = f"{DOMAIN}/film/{slug}/"
            log_id = log['data-object-id'].split(':')[-1]
            # str ^^^--- log_id: unique id of the review.
            log_link = DOMAIN + log.a['href']
            log_no = log_link.split(slug)[-1]
            log_no = int(log_no.replace('/', '')) if log_no.count('/') == 2 else 0
            # int ^^^--- log_no: there can be multiple reviews for a movie.
            #            counting starts from zero.
            #            example for first review:  /username/film/movie_name/
            #            example for first review:  /username/film/movie_name/0/
            #            example for second review: /username/film/movie_name/1/
            #                the number is specified at the end of the url ---^
            rating = log.find("span", {"class": ["rating"], })
            rating = int(rating['class'][-1].split('-')[-1]) / 2.0 if rating else None
            # float ^^^--- rating: the numerical value of the rating given in the review (0.5-5.0)
            review, spoiler = parse_review_text(log)
            # str ^^^--- review: the text content of the review.
            #            spoiler warning is checked to include or exclude the first paragraph.
            date = log.find("span", {"class": ["date"], })
            log_type = date.find_previous_sibling().text.strip()
            # str   ^^^--- log_type: Types of logs, such as:
            #              'Rewatched': (in diary) review, watched and rewatched 
            #              'Watched':   (in diary) review and watched
            #              'Added': (not in diary) review 
            date = parse_review_date(log_type, date)
            # dict ^^^--- date: the date of the review.
            #             example: {'year': 2024, 'month': 1, 'day': 1}

            data['reviews'][log_id] = {
                # static
                'movie': {
                    'name': movie_name,
                    'slug': slug,
                    'id': movie_id,
                    'release': release,
                    'link': movie_link,
                },
                # dynamic
                'type': log_type,
                'no': log_no,
                'link': log_link,
                'rating': rating,
                'review': {
                    'content': review,
                    'spoiler': spoiler
                },
                'date': date,
                'page': page,
            }

        if len(logs) < LOGS_PER_PAGE:
            data['count'] = len(data['reviews'])
            data['last_page'] = page
            break

    return data
